@@ -5,6 +5,7 @@ from __future__ import division
 from collections import OrderedDict
 import struct
 import warnings
+import zlib
 
 import numpy as np
 
@@ -26,8 +27,10 @@ class SuperBlock(object):
         version_hint = struct.unpack_from('<B', fh.peek(9), 8)[0]
         if version_hint == 0:
             contents = _unpack_struct_from_file(SUPERBLOCK_V0, fh)
+            self._end_of_sblock = offset + SUPERBLOCK_V0_SIZE
         elif version_hint == 2:
             contents = _unpack_struct_from_file(SUPERBLOCK_V2, fh)
+            self._end_of_sblock = offset + SUPERBLOCK_V2_SIZE
         else:
             raise NotImplementedError(
                 "unsupported superblock version: %i" % (version_hint))
@@ -39,7 +42,6 @@ class SuperBlock(object):
             raise NotImplementedError('File uses none 64-bit addressing')
         self.version = contents['superblock_version']
         self._contents = contents
-        self._offset = fh.tell()
         self._root_symbol_table = None
         self._fh = fh
 
@@ -47,7 +49,7 @@ class SuperBlock(object):
     def offset_to_dataobjects(self):
         """ The offset to the data objects collection for the superblock. """
         if self.version == 0:
-            sym_table = SymbolTable(self._fh, self._offset, root=True)
+            sym_table = SymbolTable(self._fh, self._end_of_sblock, root=True)
             self._root_symbol_table = sym_table
             return sym_table.group_offset
         elif self.version == 2:
@@ -63,29 +65,49 @@ class BTree(object):
 
     def __init__(self, fh, offset):
         """ initalize. """
-        fh.seek(offset)
-        self.nodes = []
-        node = _unpack_struct_from_file(B_LINK_NODE_V1, fh)
+        self.fh = fh
+
+        # read in the root node
+        root_node = self._read_node(offset)
+        self.root_node = root_node
+
+        # read in all nodes
+        all_nodes = {}
+        node_level = root_node['node_level']
+        all_nodes[node_level] = [root_node]
+        while node_level != 0:
+            new_nodes = []
+            for parent_node in all_nodes[node_level]:
+                for addr in parent_node['addresses']:
+                    new_nodes.append(self._read_node(addr))
+            new_node_level = new_nodes[0]['node_level']
+            all_nodes[new_node_level] = new_nodes
+            node_level = new_node_level
+        self.all_nodes = all_nodes
+
+    def _read_node(self, offset):
+        """ Return a single node in the B-Tree located at a given offset. """
+        self.fh.seek(offset)
+        node = _unpack_struct_from_file(B_LINK_NODE_V1, self.fh)
         assert node['signature'] == b'TREE'
 
         keys = []
         addresses = []
         for _ in range(node['entries_used']):
-            key = struct.unpack('<Q', fh.read(8))[0]
-            address = struct.unpack('<Q', fh.read(8))[0]
+            key = struct.unpack('<Q', self.fh.read(8))[0]
+            address = struct.unpack('<Q', self.fh.read(8))[0]
             keys.append(key)
             addresses.append(address)
         # N+1 key
-        keys.append(struct.unpack('<Q', fh.read(8))[0])
+        keys.append(struct.unpack('<Q', self.fh.read(8))[0])
         node['keys'] = keys
         node['addresses'] = addresses
-
-        self.nodes.append(node)
+        return node
 
     def symbol_table_addresses(self):
         """ Return a list of all symbol table address. """
         all_address = []
-        for node in self.nodes:
+        for node in self.all_nodes[0]:
             all_address.extend(node['addresses'])
         return all_address
 
@@ -145,7 +167,8 @@ class BTreeRawDataChunks(object):
         node['addresses'] = addresses
         return node
 
-    def construct_data_from_chunks(self, chunk_shape, data_shape, dtype):
+    def construct_data_from_chunks(
+            self, chunk_shape, data_shape, dtype, filter_pipeline):
         """ Build a complete data array from chunks. """
         # create array to store data
         shape = [_padded_size(i, j) for i, j in zip(data_shape, chunk_shape)]
@@ -153,11 +176,19 @@ class BTreeRawDataChunks(object):
 
         # loop over chunks reading each into the full data array
         count = np.prod(chunk_shape)
-        chunk_buffer_size = count * np.dtype(dtype).itemsize
+        itemsize = np.dtype(dtype).itemsize
+        chunk_buffer_size = count * itemsize
         for node in self.all_nodes[0]:
             for node_key, addr in zip(node['keys'], node['addresses']):
                 self.fh.seek(addr)
-                chunk_buffer = self.fh.read(chunk_buffer_size)
+                if filter_pipeline is None:
+                    chunk_buffer = self.fh.read(chunk_buffer_size)
+                else:
+                    chunk_buffer = self.fh.read(node_key['chunk_size'])
+                    filter_mask = node_key['filter_mask']
+                    chunk_buffer = self._filter_chunk(
+                        chunk_buffer, filter_mask, filter_pipeline, itemsize)
+
                 chunk_data = np.frombuffer(chunk_buffer, dtype=dtype)
                 start = node_key['chunk_offset'][:-1]
                 region = [slice(i, i+j) for i, j in zip(start, chunk_shape)]
@@ -166,6 +197,36 @@ class BTreeRawDataChunks(object):
         non_padded_region = [slice(i) for i in data_shape]
         return data[non_padded_region]
 
+    def _filter_chunk(
+            self, chunk_buffer, filter_mask, filter_pipeline, itemsize):
+        """ Apply decompression filters to a chunk of data. """
+        num_filters = len(filter_pipeline)
+        for i, pipeline_entry in enumerate(filter_pipeline[::-1]):
+
+            # A filter is skipped is the bit corresponding to its index in the
+            # pipeline is set in filter_mask
+            filter_index = num_filters - i - 1   # 0 to num_filters - 1
+            if filter_mask & (1 << filter_index):
+                continue
+
+            filter_id = pipeline_entry['filter_id']
+            if filter_id == GZIP_DEFLATE_FILTER:
+                chunk_buffer = zlib.decompress(chunk_buffer)
+
+            elif filter_id == SHUFFLE_FILTER:
+                buffer_size = len(chunk_buffer)
+                unshuffled_buffer = bytearray(buffer_size)
+                step = buffer_size // itemsize
+                for i in range(itemsize):
+                    start = i * step
+                    end = (i+1) * step
+                    unshuffled_buffer[i::itemsize] = chunk_buffer[start:end]
+                chunk_buffer = unshuffled_buffer
+            else:
+                raise NotImplementedError(
+                    "Filter with id: %i import supported" % (filter_id))
+        return chunk_buffer
+
 
 class Heap(object):
     """
@@ -288,6 +349,13 @@ class DataObjects(object):
         self._global_heaps = {}
         self._header = header
 
+        # cached attributes
+        self._filter_pipeline = None
+        self._chunk_params_set = False
+        self._chunks = None
+        self._chunk_dims = None
+        self._chunk_address = None
+
     @staticmethod
     def _parse_v1_objects(fh):
         """ Parse a collection of version 1 Data Objects. """
@@ -458,6 +526,94 @@ class DataObjects(object):
         msg_offset = msg['offset_to_message']
         return determine_dtype(self.msg_data, msg_offset)
 
+    @property
+    def chunks(self):
+        """ Tuple describing the chunk size, None if not chunked. """
+        self._get_chunk_params()
+        return self._chunks
+
+    @property
+    def compression(self):
+        """ str describing compression filter, None if no compression. """
+        if self._filter_ids is None:
+            return None
+        if GZIP_DEFLATE_FILTER in self._filter_ids:
+            return 'gzip'
+        else:
+            return None
+
+    @property
+    def compression_opts(self):
+        """ Compression filter options, None is no options/compression. """
+        if self._filter_ids is None:
+            return None
+        if GZIP_DEFLATE_FILTER in self._filter_ids:
+            gzip_entry = [d for d in self.filter_pipeline
+                          if d['filter_id'] == GZIP_DEFLATE_FILTER][0]
+            return gzip_entry['client_data'][0]
+        return None
+
+    @property
+    def shuffle(self):
+        """ Boolean indicator if shuffle filter was applied. """
+        if self._filter_ids is None:
+            return False
+        if SHUFFLE_FILTER in self._filter_ids:
+            return True
+        else:
+            return False
+
+    @property
+    def _filter_ids(self):
+        """ List of filter id in the filter pipeline, None if no pipeline. """
+        if self.filter_pipeline is None:
+            return None
+        return [d['filter_id'] for d in self.filter_pipeline]
+
+    @property
+    def filter_pipeline(self):
+        """ Dict describing filter pipeline, None if no pipeline. """
+        if self._filter_pipeline is not None:
+            return self._filter_pipeline  # use cached value
+
+        filter_msgs = self.find_msg_type(DATA_STORAGE_FILTER_PIPELINE_MSG_TYPE)
+        if len(filter_msgs) == 0:
+            self._filter_pipeline = None
+            return self._filter_pipeline
+
+        offset = filter_msgs[0]['offset_to_message']
+        version, nfilters = struct.unpack_from('<BB', self.msg_data, offset)
+        offset += struct.calcsize('<BB')
+        if version != 1:
+            raise NotImplementedError("only version 1 filters supported. ")
+
+        res0, res1 = struct.unpack_from('<HI', self.msg_data, offset)
+        offset += struct.calcsize('<HI')
+
+        filters = []
+        for _ in range(nfilters):
+            filter_info = _unpack_struct_from(
+                FILTER_PIPELINE_DESCR_V1, self.msg_data, offset)
+            offset += FILTER_PIPELINE_DESCR_V1_SIZE
+
+            padded_name_length = _padded_size(filter_info['name_length'], 8)
+            fmt = '<' + str(padded_name_length) + 's'
+            filter_name = struct.unpack_from(fmt, self.msg_data, offset)[0]
+            filter_info['filter_name'] = filter_name
+            offset += padded_name_length
+
+            fmt = '<' + str(filter_info['client_data_values']) + 'I'
+            client_data = struct.unpack_from(fmt, self.msg_data, offset)
+            filter_info['client_data'] = client_data
+            offset += 4 * filter_info['client_data_values']
+
+            if filter_info['client_data_values'] % 2:
+                offset += 4  # odd number of client data values padded
+
+            filters.append(filter_info)
+        self._filter_pipeline = filters
+        return self._filter_pipeline
+
     def get_data(self):
         """ Return the data pointed to in the DataObject. """
 
@@ -484,25 +640,40 @@ class DataObjects(object):
 
     def _get_chunked_data(self, offset):
         """ Return data which is chunked. """
+        self._get_chunk_params()
+        chunk_btree = BTreeRawDataChunks(
+            self.fh, self._chunk_address, self._chunk_dims)
+        return chunk_btree.construct_data_from_chunks(
+            self.chunks, self.shape, self.dtype, self.filter_pipeline)
+
+    def _get_chunk_params(self):
+        """
+        Get and cache chunked data storage parameters.
+
+        This method should be called prior to accessing any _chunk_*
+        attributes. Calling this method multiple times is fine, it will not
+        re-read the parameters.
+        """
+        if self._chunk_params_set:  # parameter have already need retrieved
+            return
+        self._chunk_params_set = True
+        msg = self.find_msg_type(DATA_STORAGE_MSG_TYPE)[0]
+        offset = msg['offset_to_message']
         version, layout_class = struct.unpack_from(
             '<BB', self.msg_data, offset)
         offset += struct.calcsize('<BB')
         assert version == 3
-        assert layout_class == 2
+        if layout_class != 2:  # not chunked storage
+            return
 
         dims, address = struct.unpack_from('<BQ', self.msg_data, offset)
         offset += struct.calcsize('<BQ')
-
         fmt = '<' + 'I' * (dims-1)
         chunk_shape = struct.unpack_from(fmt, self.msg_data, offset)
-        offset += struct.calcsize(fmt)
-
-        element_size = struct.unpack_from('<I', self.msg_data, offset)
-
-        chunk_btree = BTreeRawDataChunks(self.fh, address, dims)
-
-        return chunk_btree.construct_data_from_chunks(
-            chunk_shape, self.shape, self.dtype)
+        self._chunks = chunk_shape
+        self._chunk_dims = dims
+        self._chunk_address = address
+        return
 
     def find_msg_type(self, msg_type):
         """ Return a list of all messages of a given type. """
@@ -735,6 +906,7 @@ SUPERBLOCK_V0 = OrderedDict((
     ('driver_information_address', 'Q'),    # assume 8 byte addressing
 
 ))
+SUPERBLOCK_V0_SIZE = _structure_size(SUPERBLOCK_V0)
 
 # Version 2 SUPERBLOCK
 SUPERBLOCK_V2 = OrderedDict((
@@ -753,6 +925,7 @@ SUPERBLOCK_V2 = OrderedDict((
     ('superblock_checksum', 'I'),
 
 ))
+SUPERBLOCK_V2_SIZE = _structure_size(SUPERBLOCK_V2)
 
 
 B_LINK_NODE_V1 = OrderedDict((
@@ -904,6 +1077,22 @@ SYMBOL_TABLE_MSG = OrderedDict((
     ('heap_address', 'Q'),      # 8 byte addressing
 ))
 
+# IV.A.2.l The Data Storage - Filter Pipeline message
+FILTER_PIPELINE_DESCR_V1 = OrderedDict((
+    ('filter_id', 'H'),
+    ('name_length', 'H'),
+    ('flags', 'H'),
+    ('client_data_values', 'H'),
+))
+FILTER_PIPELINE_DESCR_V1_SIZE = _structure_size(FILTER_PIPELINE_DESCR_V1)
+
+RESERVED_FILTER = 0
+GZIP_DEFLATE_FILTER = 1
+SHUFFLE_FILTER = 2
+FLETCH32_FILTER = 3
+SZIP_FILTER = 4
+NBIT_FILTER = 5
+SCALEOFFSET_FILTER = 6
 
 # Data Object Message types
 # Section IV.A.2.a - IV.A.2.x
